"""Patch builtin `print` function to be thread-safe and `tqdm`-compatible."""

__all__ = ['apply']

import builtins
from functools import update_wrapper, wraps
from threading import RLock
from typing import Protocol

from ._import_hook import register_post_import_hook

_print = builtins.print
_lock = RLock()


@wraps(_print)
def locked_print(*args, **kwargs) -> None:
    with _lock:
        _print(*args, **kwargs)


def patch_print(module) -> None:
    # Create blank to force initialization of cls._lock and cls._instances
    tqdm = module.tqdm
    tqdm(disable=True)

    def tqdm_print(
        *values,
        sep: str | None = ' ',
        end: str | None = '\n',
        file: _SupportsWrite | None = None,
        flush: bool = False,
    ) -> None:
        if sep is None:
            sep = ' '
        if end is None:
            end = '\n'
        tqdm.write(sep.join(map(str, values)), end=end, file=file)

    builtins.print = update_wrapper(tqdm_print, _print)


def apply() -> None:
    builtins.print = locked_print
    register_post_import_hook(patch_print, 'tqdm')


class _SupportsWrite(Protocol):
    def write(self, s: str, /) -> object: ...



__all__ = ['wrap']

from collections.abc import Callable, Generator, Iterator
from itertools import count
from typing import Any, Protocol, Self

from wrapt import ObjectProxy

from ._types import Get


def wrap[**P, R](func: Callable[P, R], wrapper: '_Wrapper') -> Callable[P, R]:
    return _Callable(func, wrapper)


class _Wrapper(Protocol):
    @property
    def calls(self) -> count: ...

    # This one start right after function was called,
    # and stops right before it's called again.
    # Usage:
    #   fn(*args, **kwargs)
    #   resume = wrapper.suspend()
    #   ...
    #   resume()
    #   fn(*args, **kwargs)
    def suspend(self) -> Get[None]: ...

    # This one start right before function was called,
    # and stops right after it returned.
    # Usage:
    #   return wrapper(fn, *args, **kwargs)
    def __call__[**P, R](
        self, fn: Callable[P, R], /, *args: P.args, **kwds: P.kwargs
    ) -> R: ...


class _Proxy[T](ObjectProxy):  # type: ignore[misc]
    __wrapped__: T
    _self_wrapper: _Wrapper

    def __init__(self, wrapped: T, wrapper: _Wrapper) -> None:
        super().__init__(wrapped)
        self._self_wrapper = wrapper


class _Callable[**P, R](_Proxy[Callable[P, R]]):
    def __get__(
        self, instance: object, owner: type | None
    ) -> '_BoundCallable':
        fn = self.__wrapped__.__get__(instance, owner)
        return _BoundCallable(fn, self._self_wrapper)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        # patch & record fn.__call__
        next(self._self_wrapper.calls)
        r = self._self_wrapper(self.__wrapped__, *args, **kwargs)
        return _wrap(r, self._self_wrapper)


class _BoundCallable[**P, R](_Callable[P, R]):
    def __get__(self, instance: object, owner: type | None) -> Self:
        return self


# ----------------------------------- sync -----------------------------------


class _WrapStop:
    _self_wrapper: _Wrapper

    def _wrap_stop[**P, R](
        self, op: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> R:
        try:
            ret = self._self_wrapper(op, *args, **kwargs)
        except StopIteration as stop:
            stop.value = _wrap(stop.value, self._self_wrapper)
            raise
        else:
            return _wrap(ret, self._self_wrapper)


class _IterNext[Y](_WrapStop):
    __wrapped__: Iterator[Y]

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> Y:  # ! + time
        return self._wrap_stop(self.__wrapped__.__next__)


class _SendThrowClose[Y, S](_WrapStop):
    __wrapped__: Generator[Y, S, Any]

    def send(self, value: S, /) -> Y:  # ! + time
        return self._wrap_stop(self.__wrapped__.send, value)

    def throw(self, value: BaseException, /) -> Y:  # ! + time
        return self._wrap_stop(self.__wrapped__.throw, value)

    def close(self) -> Any | None:  # ! + time
        return self._wrap_stop(self.__wrapped__.close)


class _Iterator[Y](_IterNext[Y], _Proxy[Iterator[Y]]):
    pass


class _Generator[Y, S, R](
    _SendThrowClose[Y, S], _IterNext[Y], _Proxy[Generator[Y, S, R]]
):
    pass


def _wrap[T](r: T, wrapper: _Wrapper) -> T:
    # function & generator
    # are distinguishable only by their result
    match r:
        # `r` comes from generator function (types.GeneratorType)
        # or compatible
        case Generator():
            # return _gen(r, wrapper)
            return _Generator(r, wrapper)

        # `r` comes from function that returned iterator
        case Iterator():
            return _Iterator(r, wrapper)

    return r



import asyncio
import concurrent.futures as cf
from collections.abc import Callable, Hashable, Iterable, Sequence
from typing import Protocol, overload

from ._dev import hide_frame
from ._types import Coro, Some

type AnyFuture[R] = cf.Future[R] | asyncio.Future[R]
type Job[T, R] = tuple[T, AnyFuture[R]]

type BatchFn[T, R] = Callable[[Sequence[T]], Sequence[R]]
type ABatchFn[T, R] = Callable[[Sequence[T]], Coro[Sequence[R]]]


class BatchDecorator(Protocol):
    def __call__[T, R](self, fn: BatchFn[T, R], /) -> BatchFn[T, R]: ...
class ABatchDecorator(Protocol):
    def __call__[T, R](self, fn: ABatchFn[T, R], /) -> ABatchFn[T, R]: ...


class AnyBatchDecorator(Protocol):
    @overload
    def __call__[T, R](self, fn: BatchFn[T, R], /) -> BatchFn[T, R]: ...
    @overload
    def __call__[T, R](self, fn: ABatchFn[T, R], /) -> ABatchFn[T, R]: ...


def dispatch[T, R](fn: BatchFn[T, R], *xs: Job[T, R]) -> None:
    if not xs:
        return

    obj: Some[Sequence[R]] | BaseException
    try:
        with hide_frame:
            ret = fn([x for x, _ in xs])
    except BaseException as exc:  # noqa: BLE001
        obj = exc
    else:
        obj = _check_protocol(ret, len(xs))

    if isinstance(obj, Some):
        for (_, f), res in zip(xs, obj.x):
            f.set_result(res)
    else:
        for _, f in xs:
            f.set_exception(obj)


async def adispatch[T, R](fn: ABatchFn[T, R], *xs: Job[T, R]) -> None:
    if not xs:
        return

    obj: Some[Sequence[R]] | BaseException
    try:
        with hide_frame:
            ret = await fn([x for x, _ in xs])
    except asyncio.CancelledError:
        for _, f in xs:
            f.cancel()
        raise
    except BaseException as exc:  # noqa: BLE001
        obj = exc
    else:
        obj = _check_protocol(ret, len(xs))

    if isinstance(obj, Some):
        for (_, f), res in zip(xs, obj.x):
            f.set_result(res)
    else:
        for _, f in xs:
            f.set_exception(obj)
            if isinstance(f, asyncio.Future):
                f.exception()  # Mark exception as retrieved


def _check_protocol[S: Sequence](ret: S, n: int) -> Some[S] | BaseException:
    if not isinstance(ret, Sequence):
        return TypeError(
            f'Call returned non-sequence. Got {type(ret).__name__}'
        )
    if len(ret) != n:
        return RuntimeError(
            f'Call with {n} arguments '
            f'incorrectly returned {len(ret)} results'
        )
    return Some(ret)


def gather_fs[K: Hashable, R](
    fs: Iterable[tuple[K, AnyFuture[R]]],
) -> tuple[dict[K, R], BaseException | None]:
    results: dict[K, R] = {}
    errors = set[BaseException]()
    default: BaseException | None = None
    for k, f in fs:
        if f.cancelled():
            exc_tp = _fut_tp_to_cancel_tp.get(type(f))
            assert exc_tp, f'Unknown future type: {type(f).__qualname__}'
            assert default is None or isinstance(default, exc_tp)
            default = exc_tp()
        elif e := f.exception():
            errors.add(e)
        else:
            results[k] = f.result()

    match list(errors):
        case []:
            return (results, default)
        case [err]:
            return (results, err)
        case errs:
            msg = 'Got multiple exceptions'
            if all(isinstance(e, Exception) for e in errs):
                err = ExceptionGroup(msg, errs)  # type: ignore[type-var]
            else:
                err = BaseExceptionGroup(msg, errs)
            return (results, err)


_fut_tp_to_cancel_tp: dict[type[AnyFuture], type[BaseException]] = {
    cf.Future: cf.CancelledError,
    asyncio.Future: asyncio.CancelledError,
}



from argparse import ArgumentParser
from collections.abc import Callable, Mapping, Sequence
from typing import Any, overload

from ._types import Get

@overload
def arg[T](
    default: T,
    /,
    *,
    flag: str = ...,
    init: bool = ...,
    repr: bool = ...,
    hash: bool = ...,
    help: str = ...,
    compare: bool = ...,
    metadata: Mapping[str, object] = ...,
) -> T: ...
@overload
def arg[T](
    *,
    factory: Get[T],
    flag: str = ...,
    init: bool = ...,
    repr: bool = ...,
    hash: bool = ...,
    help: str = ...,
    compare: bool = ...,
    metadata: Mapping[str, object] = ...,
) -> T: ...
@overload
def arg(
    *,
    flag: str = ...,
    init: bool = ...,
    repr: bool = ...,
    hash: bool = ...,
    help: str = ...,
    compare: bool = ...,
    metadata: Mapping[str, object] = ...,
) -> Any: ...
def parse_args[T](
    fn: Callable[..., T], args: Sequence[str] = ..., prog: str = ...
) -> tuple[T, ArgumentParser]: ...



"""Make builtin iterators compatible with `len`.

Supports:
- str, bytes, range, tuple, list, set, deque iterators
- dict keys/values/items iterators
- OrderedDict keys/values/items iterators
- reversed
- map
- zip

Since Python 3.12 `itertools` doesn't support serialization,
thus `len` will never work for them.
"""

__all__ = ['apply']

import builtins
import functools
import operator
from collections import OrderedDict, deque
from collections.abc import Iterable

# --------------------------------- builtins ---------------------------------

len_hint = functools.singledispatch(builtins.len)

_iterables: list[Iterable] = [
    '',
    b'',
    range(0),
    (),
    [],
    {},
    {}.keys(),
    {}.values(),
    {}.items(),
    reversed(()),
    reversed([]),
    set(),
    frozenset(),
    deque(),
]
_transparent_types: tuple[type, ...] = tuple(
    it.__iter__().__class__ for it in _iterables
)
for _tp in _transparent_types:
    len_hint.register(_tp, operator.length_hint)

_odict_iter_tp: type = OrderedDict().__iter__().__class__


def _are_definitely_independent(iters) -> bool:
    return len({id(it) for it in iters}) == len(iters) and all(
        isinstance(it, _transparent_types) for it in iters
    )


@len_hint.register(zip)
def _len_zip(x) -> int:  # type: ignore[misc]
    _, iters = x.__reduce__()
    if not iters:
        return 0
    if len(iters) == 1:
        return len(iters[0])

    # Do not compute zip size when it's constructed from multiple iterables.
    # as there's currently no reliable way to check whether underlying
    # iterables are independent or not
    if _are_definitely_independent(iters):
        return min(map(len, iters))

    raise TypeError


@len_hint.register(map)
def _len_map(x) -> int:  # type: ignore[misc]
    _, (__fn, *iters) = x.__reduce__()
    if len(iters) == 1:
        return len(iters[0])

    # Same as for zip above
    if _are_definitely_independent(iters):
        return min(map(len, iters))

    raise TypeError


@len_hint.register(_odict_iter_tp)
def _len_odict_iter(x) -> int:
    _, [items] = x.__reduce__()
    return len(items)


def apply() -> None:
    builtins.len = len_hint



from collections.abc import Callable
from contextlib import AbstractContextManager
from typing import overload

from ._futures import BatchDecorator, BatchFn
from ._types import Get

def threadlocal[T, **P](
    fn: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
) -> Get[T]: ...
def interpreter_lock(timeout: float = ...) -> AbstractContextManager[None]: ...
def call_once[T](fn: Get[T], /) -> Get[T]: ...
def shared_call[**P, R](fn: Callable[P, R], /) -> Callable[P, R]: ...
def weak_memoize[**P, R](fn: Callable[P, R], /) -> Callable[P, R]: ...
@overload
def streaming(
    *,
    batch_size: int | None = ...,
    timeout: float = ...,
    workers: int = ...,
    pool_timeout: float = ...,
) -> BatchDecorator: ...
@overload
def streaming[T, R](
    func: BatchFn[T, R],
    /,
    *,
    batch_size: int | None = ...,
    timeout: float = ...,
    workers: int = ...,
    pool_timeout: float = ...,
) -> BatchFn[T, R]: ...



__all__ = ['amap', 'amap_dict', 'astarmap', 'azip']

import asyncio
from asyncio import Queue, Task
from collections import deque
from collections.abc import (
    AsyncIterator,
    Callable,
    Collection,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
)
from contextlib import suppress
from functools import partial
from typing import TypeGuard, cast, overload

from ._dev import hide_frame
from ._futures import ABatchDecorator, ABatchFn, Job, adispatch
from ._types import AnyIterable, AnyIterator, Coro


async def amap_dict[K, T1, T2](
    func: Callable[[T1], Coro[T2]],
    obj: Mapping[K, T1],
    /,
    *,
    limit: int,
) -> dict[K, T2]:
    """Asynchronously apply `func` to each value in a mapping.

    For extra options, see astarmap, which is used under hood.
    """
    aiter_values = amap(func, obj.values(), limit=limit)
    values = [v async for v in aiter_values]
    return dict(zip(obj.keys(), values, strict=True))


def amap[R](
    func: Callable[..., Coro[R]],
    /,
    *iterables: AnyIterable,
    limit: int,
    unordered: bool = False,
) -> AsyncIterator[R]:
    """Async version of map(func, *iterables).

    Make an iterator that computes the function using arguments from
    each of the iterables. Stops when the shortest iterable is exhausted.

    For extra options, see `astarmap`.
    """
    it = zip(*iterables) if _all_sync_iters(iterables) else azip(*iterables)
    return astarmap(func, it, limit=limit, unordered=unordered)


async def astarmap[*Ts, R](
    func: Callable[[*Ts], Coro[R]],
    iterable: AnyIterable[tuple[*Ts]],
    /,
    *,
    limit: int,
    unordered: bool = False,
) -> AsyncIterator[R]:
    """Async version of itertools.starmap(fn, iterable).

    Return an iterator whose values are returned from the function evaluated
    with an argument tuple taken from the given sequence.

    Options:
    - limit - Maximum number of simultaneously running tasks.
    - unordered - Set to get yield results as soon as they are ready.
    """
    assert callable(func)

    # optimization: Plain loop if concurrency is unnecessary
    if limit <= 1:
        if isinstance(iterable, Iterable):
            for args in iterable:
                yield await func(*args)
        else:
            async for args in iterable:
                yield await func(*args)
        return

    async with asyncio.TaskGroup() as tg:
        ts = (
            (tg.create_task(func(*args)) for args in iterable)
            if isinstance(iterable, Iterable)
            else (tg.create_task(func(*args)) async for args in iterable)
        )

        it = (
            _iter_results_unordered(ts, limit=limit)
            if unordered
            else _iter_results(ts, limit=limit)
        )

        async for x in it:
            yield x


async def _iter_results_unordered[T](
    ts: AnyIterator[Task[T]], limit: int
) -> AsyncIterator[T]:
    """Fetch and run async tasks.

    Runs exactly `limit` tasks simultaneously (less in the end of iteration).
    Order of results is arbitrary.
    """
    todo = set[Task[T]]()
    done = Queue[Task[T]]()

    def _todo_to_done(t: Task[T]) -> None:
        todo.discard(t)
        done.put_nowait(t)

    while True:
        # Prefill task buffer
        while len(todo) + done.qsize() < limit and (
            t := (
                next(ts, None)
                if isinstance(ts, Iterator)
                else await anext(ts, None)
            )
        ):
            # optimization: Immediately put to done if the task is
            # already done (e.g. if the coro was able to complete eagerly),
            # and skip scheduling a done callback
            if t.done():
                done.put_nowait(t)
            else:
                todo.add(t)
                t.add_done_callback(_todo_to_done)

        # No more tasks to do and nothing more to schedule
        if not todo and done.empty():
            return

        # Wait till any task succeed
        yield (await done.get()).result()

        # Pop tasks happened to also be DONE (after line above)
        while not done.empty():
            yield done.get_nowait().result()


async def _iter_results[T](
    ts: AnyIterator[Task[T]], limit: int
) -> AsyncIterator[T]:
    """Fetch and run async tasks.

    Runs up to `limit` tasks simultaneously (less in the end of iteration).
    Order of results is preserved.
    """
    todo = deque[Task[T]]()
    while True:
        # Prefill task buffer
        while len(todo) < limit and (
            t := (
                next(ts, None)
                if isinstance(ts, Iterator)
                else await anext(ts, None)
            )
        ):
            todo.append(t)
        if not todo:  # No more tasks to do and nothing more to schedule
            return

        # Forcefully block first task, while it's awaited,
        # others in `todo` are also running, because they are `asyncio.Task`.
        # So after this some of tasks from `todo` are also done.
        yield await todo.popleft()

        # Pop tasks happened to also be DONE (after line above)
        while todo and todo[0].done():
            yield todo.popleft().result()


async def azip(*iterables: AnyIterable) -> AsyncIterator[tuple]:
    if _all_sync_iters(iterables):
        for x in zip(*iterables):
            yield x
        return

    aiters = (
        _wrapgen(it) if isinstance(it, Iterable) else aiter(it)
        for it in iterables
    )
    while True:
        try:
            ret = await asyncio.gather(*(anext(ait) for ait in aiters))
        except StopAsyncIteration:
            return
        else:
            yield tuple(ret)


def _all_sync_iters(
    iterables: Collection[AnyIterable],
) -> TypeGuard[Collection[Iterable]]:
    return all(isinstance(it, Iterable) for it in iterables)


async def _wrapgen[T](it: Iterable[T]) -> AsyncIterator[T]:
    for x in it:
        yield x


@overload
def astreaming(
    *, batch_size: int | None = ..., timeout: float = ...
) -> ABatchDecorator: ...


@overload
def astreaming[T, R](
    fn: ABatchFn[T, R],
    /,
    *,
    batch_size: int | None = ...,
    timeout: float = ...,
) -> ABatchFn[T, R]: ...


def astreaming[T, R](
    fn: ABatchFn[T, R] | None = None,
    /,
    *,
    batch_size: int | None = None,
    timeout: float = 0.1,
) -> ABatchFn[T, R] | ABatchDecorator:
    """Compute on `timeout` or if batch is collected.

    `timeout` (in seconds) is a time to wait till the batch is full,
    i.e. latency.

    Uses ideas from
    - https://github.com/ShannonAI/service-streamer
    - https://github.com/leon0707/batch_processor
    - ray.serve.batch
      https://github.com/ray-project/ray/blob/master/python/ray/serve/batching.py

    Note: currently supports only functions and bound methods.

    Implementation details:
    - any caller enqueues jobs and starts waiting
    """
    if fn is None:
        deco = partial(astreaming, batch_size=batch_size, timeout=timeout)
        return cast('ABatchDecorator', deco)

    assert batch_size is None or batch_size >= 1
    assert timeout > 0

    buf: list[Job[T, R]] = []
    deadline = float('-inf')
    not_last = asyncio.Event()
    lock = asyncio.Lock()
    ncalls = 0

    async def wrapper(items: Sequence[T]) -> list[R]:
        nonlocal ncalls, deadline
        if not items:
            return []

        # There's another handling call with tail, wake it up
        if not ncalls and buf:
            not_last.set()

        ncalls += 1
        fs: list[asyncio.Future[R]] = []
        try:
            for x in items:
                f = asyncio.Future[R]()
                fs.append(f)
                buf.append((x, f))

                if len(buf) == 1:  # Got first job, reset deadline
                    deadline = asyncio.get_running_loop().time() + timeout

                # Full batch, dispatch
                if batch_size is not None and len(buf) == batch_size:
                    batch, buf[:] = buf[:], []
                    async with lock:
                        await adispatch(fn, *batch)
        finally:
            ncalls -= 1

        if not ncalls and buf:  # Was last call, wait for another
            not_last.clear()

            notified = False
            with suppress(TimeoutError):
                async with asyncio.timeout_at(deadline):
                    notified = await not_last.wait()

            if not notified:
                batch, buf[:] = buf[:], []
                async with lock:
                    await adispatch(fn, *batch)

        with hide_frame:
            return await asyncio.gather(*fs)

    return wrapper



__all__ = ['Sound']

from contextlib import ExitStack
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from queue import Queue
from threading import Event

import numpy as np
import numpy.typing as npt
from tqdm.auto import tqdm

from .. import chunked

try:
    from sounddevice import CallbackAbort, CallbackStop, OutputStream
except ImportError:
    OutputStream = None
    CallbackAbort = CallbackStop = Exception

try:
    import soundfile
except ImportError:
    soundfile = None


def _play(
    arr: np.ndarray, rate: int, blocksize: int = 1024, bufsize: int = 20
) -> None:
    """Play audio from array. Crtl-C to interrupt."""
    if OutputStream is None:
        raise ImportError('Cannot import `sounddevice` module')

    q: Queue[np.ndarray | None] = Queue(bufsize)
    ev = Event()

    def callback(out: np.ndarray, *_) -> None:
        if (data := q.get()) is None:
            raise CallbackAbort

        size = len(data)
        out[:size] = data
        if size < len(out):
            out[size:] = 0
            raise CallbackStop

    stream = OutputStream(
        rate, blocksize, callback=callback, finished_callback=ev.set
    )

    fmt = '{percentage:3.0f}% |{bar}| [{elapsed}<{remaining}]'
    blocks = chunked(arr, blocksize)

    with ExitStack() as s:
        s.enter_context(stream)  # Close stream
        s.callback(ev.wait)  # Wait for completion
        s.callback(q.put, None)  # Close queue

        for data in s.enter_context(
            tqdm(blocks, leave=False, smoothing=0, bar_format=fmt)
        ):
            q.put(data)


@dataclass(repr=False, frozen=True)
class Sound[S: np.number]:
    """Wrap numpy.array to be playable as sound.

    Parameters:
    - rate - sample rate to use for playback

    Usage:
    ```
    import numpy as np

    sound = Sound.load('test.flac')

    # Get properties
    rate: int = sound.rate
    dtype: np.dtype = sound.dtype

    # Could be played like:
    import sounddevice as sd
    sd.play(sound, sound.rate)

    # Or like this, if you need Ctrl-C support
    sound.play()

    # Extract underlying array
    raw = sound.raw

    # Same result
    raw = np.array(sound)
    ```
    """

    data: npt.NDArray[S]
    rate: int = 44_100

    def __post_init__(self) -> None:
        if self.data.ndim not in (1, 2):
            msg = f'Sound must be 1d (mono) or 2d array, got {self.data.shape}'
            raise ValueError(msg)
        if self.data.ndim == 1:
            object.__setattr__(self, 'data', self.data[:, None])
        if self.data.shape[-1] not in (1, 2):
            msg = (
                f'Only mono/stereo is supported, got {self.channels} channels'
            )
            raise ValueError(msg)
        if self.data.dtype not in ('i1', 'i2', 'i4', 'f4'):
            msg = (
                'Only int8/int16/int32/float32 sound dtype is supported. '
                f'Got {self.data.dtype}'
            )
            raise ValueError(msg)

    @property
    def channels(self) -> int:
        return self.data.shape[1]

    @property
    def duration(self) -> timedelta:
        return timedelta(seconds=self.data.shape[0] / self.rate)

    def __repr__(self) -> str:
        duration = self.duration
        channels = self.channels
        dtype = self.data.dtype
        return f'{type(self).__name__}({duration=!s}, {channels=}, {dtype=!s})'

    def __array__(self) -> npt.NDArray[S]:
        return self.data

    def play(self, blocksize=1024) -> None:
        """Play audio from array. Supports interruption via Crtl-C."""
        _play(self.data, self.rate, blocksize=blocksize)

    @classmethod
    def load(cls, path: Path | str) -> 'Sound':
        if soundfile is None:
            raise ImportError('Cannot import `soundfile` module')
        _check_fmt(path)
        data, rate = soundfile.read(path)
        return cls(data.astype('f'), rate)

    def save(self, path: Path | str) -> None:
        if soundfile is None:
            raise ImportError('Cannot import `soundfile` module')
        _check_fmt(path)
        soundfile.write(path, self.data, self.rate)


def _check_fmt(path: Path | str) -> None:
    fmt = Path(path).suffix.lower()
    if fmt not in ('.flac', '.wav'):
        msg = f'Only FLAC/WAV is supported, got {fmt}'
        raise ValueError(msg)



__all__ = [
    'buffered',
    'get_executor',
    'map_n',
    'map_n_dict',
    'max_cpu_count',
    'starmap_n',
]

import atexit
import enum
import os
import signal
import sys
import warnings
import weakref
from collections.abc import Callable, Iterable, Iterator, Mapping, Sized
from concurrent.futures import Executor, Future
from contextlib import ExitStack, contextmanager
from cProfile import Profile
from functools import partial
from itertools import chain, islice, starmap
from logging import getLogger
from multiprocessing.managers import BaseManager
from operator import methodcaller
from pstats import Stats
from queue import Empty, SimpleQueue
from threading import Lock
from time import monotonic, sleep
from typing import Final, Protocol, Self, cast

import loky

try:
    import psutil
except ImportError:
    psutil = None

from ._dev import hide_frame
from ._more import chunked, ilen
from ._reduction import move_to_shmem, reducers
from ._thread_quota import ThreadQuota
from ._types import Get, Some

_TOTAL_CPUS = (
    os.process_cpu_count() if sys.version_info >= (3, 13) else os.cpu_count()
)
_NUM_CPUS = _TOTAL_CPUS or 0

if (_env_cpus := os.getenv('GLOW_CPUS')) is not None:
    _NUM_CPUS = min(_NUM_CPUS, int(_env_cpus))
    _NUM_CPUS = max(_NUM_CPUS, 0)
_IDLE_WORKER_TIMEOUT = 10
_GRANULAR_SCHEDULING = False  # TODO: investigate whether this improves load
_LOGGER = getLogger(__name__)


class _Empty(enum.Enum):
    token = 0


_empty: Final = _Empty.token

# ------------------- some useful interfaces and functions -------------------


class _Queue[T](Protocol):
    def get(self, block: bool = ..., timeout: float | None = ...) -> T: ...
    def put(self, item: T) -> None: ...


class _Event(Protocol):
    def is_set(self) -> bool: ...
    def set(self) -> None: ...


class _Manager(Protocol):
    def Event(self) -> _Event: ...  # noqa: N802
    def Queue(self, /, maxsize: int) -> _Queue: ...  # noqa: N802


def _get_cpu_count_limits(
    upper_bound: int = sys.maxsize, *, mp: bool = False
) -> Iterator[int]:
    yield from (upper_bound, _TOTAL_CPUS or 1)

    # Windows platform lacks memory overcommit, so it's sensitive to VMS growth
    if not mp or sys.platform != 'win32' or 'torch' not in sys.modules:
        return

    if (sys.modules['torch'].version.cuda or '') >= '11.7.0':
        # It's expected that torch will fix .nv_fatb readonly flag in its DLLs
        # See https://stackoverflow.com/a/69489193/9868257
        return

    if psutil is None:
        warnings.warn(
            'Max process count may be calculated incorrectly, '
            'leading to application crash or even BSOD. '
            'Install psutil to avoid that',
            stacklevel=3,
        )
        return

    # Overcommit on Windows is forbidden, thus VMS planning is necessary
    vms: int = psutil.Process().memory_info().vms
    free_vms: int = psutil.virtual_memory().free + psutil.swap_memory().free
    yield free_vms // vms


def max_cpu_count(upper_bound: int = sys.maxsize, *, mp: bool = False) -> int:
    return min(_get_cpu_count_limits(upper_bound, mp=mp))


_PATIENCE = 0.01


class _TimeoutCallable[T](Protocol):
    def __call__(self, *, timeout: float) -> T: ...


def _retry_call[T](fn: _TimeoutCallable[T], *exc: type[BaseException]) -> T:
    # See issues
    # https://github.com/dask/dask/pull/2144#issuecomment-290556996
    # https://github.com/dask/dask/pull/2144/files
    # https://github.com/python/cpython/issues/74157
    # FIXED in py3.15+
    while True:
        try:
            return fn(timeout=_PATIENCE)
        except exc:
            sleep(0)  # Force switch to another thread to proceed


if sys.platform == 'win32':

    def _exception[T](f: Future[T], /) -> BaseException | None:
        return _retry_call(f.exception, TimeoutError)

else:
    _exception = Future.exception


def _result[T](f: Future[T], cancel: bool = True) -> Some[T] | BaseException:
    try:
        return exc if (exc := _exception(f)) else Some(f.result())
    finally:
        if cancel:
            f.cancel()
        del f


def _q_get_fn[T](q: _Queue[T]) -> Get[T]:
    if sys.platform != 'win32':
        return q.get
    return partial(_retry_call, q.get, Empty)


# ---------------------------- pool initialization ----------------------------


def _mp_profile() -> None:
    """Multiprocessed profiler."""
    prof = Profile()
    prof.enable()

    def _finalize(lines=50) -> None:
        prof.disable()
        with open(f'prof-{os.getpid()}.txt', 'w') as fp:
            Stats(prof, stream=fp).sort_stats('cumulative').print_stats(lines)

    atexit.register(_finalize)


def _initializer() -> None:
    # `signal.signal` suppresses KeyboardInterrupt in child processes
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    if os.environ.get('_GLOW_MP_PROFILE'):
        _mp_profile()


@contextmanager
def get_executor(max_workers: int, *, mp: bool) -> Iterator[Executor]:
    if mp:
        processes: loky.ProcessPoolExecutor = loky.get_reusable_executor(
            max_workers,
            'loky_init_main',
            _IDLE_WORKER_TIMEOUT,
            job_reducers=reducers,
            result_reducers=reducers,
            initializer=_initializer,
        )
        # In generator 'finally' is not reliable enough, use atexit
        hook = atexit.register(processes.shutdown, kill_workers=True)
        yield processes
        atexit.unregister(hook)
    else:
        threads = ThreadQuota(max_workers)
        try:
            yield threads
        finally:
            is_success = sys.exception() is None
            threads.shutdown(wait=is_success, cancel_futures=True)


def _get_manager(executor: Executor) -> _Manager:
    if isinstance(executor, loky.ProcessPoolExecutor):
        return executor._context.Manager()

    else:  # noqa: RET505
        from multiprocessing.dummy import Manager

        return Manager()


# -------- bufferize iterable by offloading to another thread/process --------


def _consume[T](
    items: Iterable[T], buf: _Queue[T | _Empty], stop: _Event
) -> None:
    try:
        for item in items:
            if stop.is_set():
                break
            buf.put(item)
    finally:
        buf.put(_empty)  # Signal to stop iteration
        buf.put(_empty)  # Match last q.get


class buffered[T](Iterator[T]):  # noqa: N801
    """Iterate in background thread with at most `latency` items ahead."""

    __slots__ = ('__weakref__', '_consume', '_next', 'close')

    def __init__(
        self,
        iterable: Iterable[T],
        /,
        *,
        latency: int = 2,
        mp: bool | Executor = False,
    ) -> None:
        s = ExitStack()
        if isinstance(mp, Executor):
            executor = mp
        else:
            executor = s.enter_context(get_executor(1, mp=mp))

        mgr = _get_manager(executor)
        if isinstance(mgr, BaseManager):
            s.enter_context(mgr)

        ev: _Event = mgr.Event()
        q: _Queue[T | _Empty] = mgr.Queue(latency)
        self._consume = executor.submit(_consume, iterable, q, ev)
        self._next = _q_get_fn(q)

        # If main is killed, unblocks consumer to allow it to check stop flag
        # Otherwise collects 2nd _empty from q.
        # Called 2nd
        s.callback(self._next)

        # If main is killed, notifies consumer to stop.
        # If consumer is already stopped (on error or not), does nothing.
        # Called 1st
        s.callback(ev.set)

        self.close = weakref.finalize(self, s.close)

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> T:
        if self.close.alive:
            if (item := self._next()) is not _empty:
                return item

            self.close()
            # Reraise exception from source iterable if any
            obj = _result(self._consume, cancel=False)
            if not isinstance(obj, Some):
                with hide_frame:
                    raise obj

        raise StopIteration


# ---------------------------- automatic batching ----------------------------


class _AutoSize:
    MIN_DURATION = 0.2
    MAX_DURATION = 2.0
    size: int = 1
    duration: float = 0.0

    def __init__(self) -> None:
        self.lock = Lock()
        assert self.MIN_DURATION * 2 < self.MAX_DURATION, 'Range is too tight'

    def suggest(self) -> int:
        with self.lock:
            return self.size

    def update(self, n: int, start_time: float, fut: Future[Sized]) -> None:
        # Compute as soon as future became done, discard later if not needed
        duration = monotonic() - start_time

        if fut.cancelled():  # Job never run, zero load
            return

        with self.lock:
            if n != self.size:  # Ran with old size
                return

            # Do EMA smoothing
            self.duration = (
                (0.8 * self.duration + 0.2 * duration)
                if self.duration > 0
                else duration
            )
            if self.duration <= 0:  # Smh not initialized yet
                return  # Or duration is less then `monotonic()` precision

            if self.duration < self.MIN_DURATION:  # Too high IPC overhead
                size = self.size * 2
                _LOGGER.debug('Doubling batch size to %d', size)

            elif (
                self.duration <= self.MAX_DURATION  # Range is optimal
                or self.size == 1  # Cannot reduce already minimal batch
            ):
                return

            else:  # Too high latency
                size = int(2 * self.size * self.MIN_DURATION / self.duration)
                size = max(size, 1)
                _LOGGER.debug('Reducing batch size to %d', size)

            self.size = size
            self.duration = 0.0


# ---------------------- map iterable through function ----------------------


def _schedule[F: Future](
    make_future: Callable[..., F], args_zip: Iterable[Iterable], chunksize: int
) -> Iterator[F]:
    return starmap(make_future, chunked(args_zip, chunksize))


def _schedule_auto[F: Future](
    make_future: Callable[..., F],
    args_zip: Iterator[Iterable],
    max_workers: int,
) -> Iterator[F]:
    # For the whole wave make futures with the same job size
    size = _AutoSize()
    while tuples := [*islice(args_zip, size.suggest() * max_workers)]:
        chunksize = len(tuples) // max_workers or 1
        for args in chunked(tuples, chunksize):
            f = make_future(*args)
            f.add_done_callback(partial(size.update, len(args), monotonic()))
            yield f


def _schedule_auto_v2[F: Future](
    make_future: Callable[..., F], args_zip: Iterator[Iterable]
) -> Iterator[F]:
    # Vary job size from future to future
    size = _AutoSize()
    while args := [*islice(args_zip, size.suggest())]:
        f = make_future(*args)
        f.add_done_callback(partial(size.update, len(args), monotonic()))
        yield f


def _get_unwrap_iter[T](
    s: ExitStack,
    qsize: int,
    get_done_f: Get[Future[T]],
    fs_scheduler: Iterator,
) -> Iterator[T]:
    with s:
        if not qsize:  # No tasks to do
            return

        # Unwrap 1st / schedule `N-qsize` / unwrap `qsize-1`
        with hide_frame:
            for _ in chain([None], fs_scheduler, range(qsize - 1)):
                # Retrieve done task, exactly `N` calls
                obj = _result(get_done_f())
                if not isinstance(obj, Some):
                    with hide_frame:
                        raise obj
                yield obj.x


def _unwrap[T](
    s: ExitStack,
    fs: Iterable[Future[T]],
    *,
    qsize: int | None,
    order: bool,
) -> Iterator[T]:
    q = SimpleQueue[Future[T]]()

    # If `order`, then `q` has "PENDING"/"RUNNING"/"DONE" tasks,
    # otherwise it only has "DONE" tasks.
    # FIXME: order=False -> random freezes (in q.get -> Empty)
    q_put = cast(
        'Callable[[Future[T]], None]',
        q.put if order else methodcaller('add_done_callback', q.put),
    )

    # On each `next()` schedules new task
    fs_scheduler = map(q_put, fs)
    try:
        # Fetch up to `qsize` tasks to pre-fill `q`
        qsize = ilen(islice(fs_scheduler, qsize))

    except BaseException:
        # Unwind stack here on an error
        s.close()
        raise

    else:
        return _get_unwrap_iter(s, qsize, _q_get_fn(q), fs_scheduler)


def _batch_invoke[*Ts, R](
    func: Callable[[*Ts], R], *items: tuple[*Ts]
) -> list[R]:
    return [*starmap(func, items)]


def starmap_n[T](
    func: Callable[..., T],
    iterable: Iterable[Iterable],
    /,
    *,
    max_workers: int | None = None,
    prefetch: int | None = 2,
    mp: bool = False,
    chunksize: int | None = None,
    order: bool = True,
) -> Iterator[T]:
    """Equivalent to itertools.starmap(fn, iterable).

    Return an iterator whose values are returned from the function evaluated
    with an argument tuple taken from the given sequence.

    Options:
    - workers - Count of workers, by default all hardware threads are occupied.
    - prefetch - Extra count of scheduled jobs, if not set equals to infinity.
    - mp - Whether use processes or threads.
    - chunksize - The size of the chunks the iterable will be broken into
      before being passed to a processes. Estimated automatically.
      Ignored when threads are used.
    - order - Whether keep results order, or ignore it to increase performance.

    Unlike multiprocessing.Pool or concurrent.futures.Executor this one:
    - never deadlocks on any exception or Ctrl-C interruption.
    - accepts infinite iterables due to lazy task creation (option prefetch).
    - has single interface for both threads and processes.
    - TODO: serializes array-like data using out-of-band Pickle 5 buffers.
    - before first `__next__` call it submits at most `prefetch` jobs
      to warmup pool of workers.

    Notes:
    - To reduce latency set order to False, order of results will be arbitrary.
    - To increase CPU usage increase prefetch or set it to None.
    - In terms of CPU usage there's no difference between
      prefetch=None and order=False, so choose wisely.
    - Setting order to False makes no use of prefetch more than 0.

    TODO: replace `order=True` with `heap=False`
    """
    if max_workers is None:
        max_workers = max_cpu_count(_NUM_CPUS, mp=mp)

    if not max_workers or not _NUM_CPUS:
        return starmap(func, iterable)  # Fallback to single thread

    if mp and chunksize is None and prefetch is None:
        msg = 'With multiprocessing either chunksize or prefetch should be set'
        raise ValueError(msg)

    if prefetch is not None:
        prefetch = max(prefetch + max_workers, 1)

    it = iter(iterable)
    s = ExitStack()
    submit = s.enter_context(get_executor(max_workers, mp=mp)).submit

    if mp:
        func = move_to_shmem(func)
    else:
        chunksize = chunksize or 1

    if chunksize == 1:
        submit_one = cast('Callable[..., Future[T]]', partial(submit, func))
        f1s = starmap(submit_one, it)
        return _unwrap(s, f1s, qsize=prefetch, order=order)

    submit_many = cast(
        'Callable[..., Future[list[T]]]', partial(submit, _batch_invoke, func)
    )
    if chunksize is not None:
        # Fixed chunksize
        fs = _schedule(submit_many, it, chunksize)
    elif not _GRANULAR_SCHEDULING:
        # Dynamic chunksize scaling, submit tasks in waves
        fs = _schedule_auto(submit_many, it, max_workers)
    else:
        # Dynamic chunksize scaling
        fs = _schedule_auto_v2(submit_many, it)

    chunks = _unwrap(s, fs, qsize=prefetch, order=order)
    return chain.from_iterable(chunks)


def map_n[T](
    func: Callable[..., T],
    /,
    *iterables: Iterable,
    max_workers: int | None = None,
    prefetch: int | None = 2,
    mp: bool = False,
    chunksize: int | None = None,
    order: bool = True,
) -> Iterator[T]:
    """Return iterator equivalent to map(func, *iterables).

    Make an iterator that computes the function using arguments from
    each of the iterables. Stops when the shortest iterable is exhausted.

    For extra options, see starmap_n, which is used under hood.
    """
    return starmap_n(
        func,
        zip(*iterables),
        max_workers=max_workers,
        prefetch=prefetch,
        mp=mp,
        chunksize=chunksize,
        order=order,
    )


def map_n_dict[K, T1, T2](
    func: Callable[[T1], T2],
    obj: Mapping[K, T1],
    /,
    *,
    max_workers: int | None = None,
    prefetch: int | None = 2,
    mp: bool = False,
    chunksize: int | None = None,
) -> dict[K, T2]:
    """Apply `func` to each value in a mapping in parallel way.

    For extra options, see starmap_n, which is used under hood.
    """
    iter_values = map_n(
        func,
        obj.values(),
        max_workers=max_workers,
        prefetch=prefetch,
        mp=mp,
        chunksize=chunksize,
    )
    return dict(zip(obj.keys(), iter_values, strict=True))



__all__ = ['init_loguru']

import logging
import sys
from collections.abc import Iterable
from types import FrameType, ModuleType
from typing import TYPE_CHECKING, Any, TypedDict, Unpack

from loguru import logger

if TYPE_CHECKING:
    from loguru import FilterDict, FilterFunction

_DEFAULT_FMT = (
    '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>'
    ' | '
    '<level>{level: <8}</level>'
    ' | '
    '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>'
    ' | '
    '<level>{message}</level>'
)


class _LoggerAddKwds(TypedDict, total=False):
    colorize: bool | None
    serialize: bool
    backtrace: bool
    diagnose: bool
    filter: 'str | FilterFunction | FilterDict'


def init_loguru(
    level: str = 'WARNING',
    *,
    names: Iterable[str] | dict[str, list[str]] = (),
    fmt: str = _DEFAULT_FMT,
    **logger_add_kwargs: Unpack[_LoggerAddKwds],
) -> None:
    """
    Configure loguru.

    Does:
    - remap all `logging.Logger` to `loguru` calls
    - remap all warnings to `logger.warning`
    - configure `uvicorn`, `hypercorn` and `websockets` loggers.
    """
    logging.basicConfig(
        level=level,
        handlers=[_InterceptHandler()],
        force=True,
    )
    logging.captureWarnings(True)

    logger.remove()
    logger.add(sys.stdout, level=level, format=fmt, **logger_add_kwargs)
    _intercept_std_logger('', level)

    if not isinstance(names, dict):
        names = {level: [*names]}

    names[level] = [
        'uvicorn',
        'uvicorn.access',
        'uvicorn.error',
        'hypercorn',
        'hypercorn.access',
        'hypercorn.error',
        'sse_starlette.sse',
        'websockets',
        'websockets.client',
        'websockets.protocol',
        'websockets.server',
        *names.get(level, []),
    ]
    for level_, names_ in names.items():
        for name in names_:
            _intercept_std_logger(name, level_)


def _intercept_std_logger(module: str | ModuleType, level: int | str) -> None:
    log = logging.getLogger(
        module if isinstance(module, str) else module.__name__
    )
    log.handlers = [_InterceptHandler(level=level)]
    log.propagate = False


class _InterceptHandler(logging.Handler):
    def emit(self, record: Any) -> None:
        # Get corresponding Loguru level if it exists.
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame: FrameType | None = logging.currentframe()
        depth = 0
        while frame and 'logging' in frame.f_code.co_filename:
            frame = frame.f_back
            depth += 1

        opt = logger.opt(exception=record.exc_info, depth=depth)
        opt.log(level, record.getMessage())



__all__ = ['register_post_import_hook', 'when_imported']

import sys
from collections.abc import Callable, Sequence
from importlib import abc, util
from importlib.machinery import ModuleSpec
from threading import RLock
from types import ModuleType

from ._types import Callback

_INITIALIZED = False
_LOCK = RLock()
_HOOKS: dict[str, list[Callback[ModuleType]]] = {}


class _ImportHookChainedLoader(abc.Loader):
    def __init__(self, loader: abc.Loader) -> None:
        self.loader = loader

    def _set_loader(self, module: ModuleType) -> None:
        undefined = object()
        if getattr(module, '__loader__', undefined) in (None, self):
            try:
                module.__loader__ = self.loader
            except AttributeError:
                pass

        if (spec := getattr(module, '__spec__', None)) is not None and getattr(
            spec, 'loader', None
        ) is self:
            spec.loader = self.loader

    def create_module(self, spec: ModuleSpec) -> ModuleType | None:
        return self.loader.create_module(spec)

    def exec_module(self, module: ModuleType) -> None:
        self._set_loader(module)
        self.loader.exec_module(module)

        name: str | None = getattr(module, '__name__', None)
        with _LOCK:
            hooks = _HOOKS.pop(name, [])  # type: ignore[arg-type]
        for hook in hooks:
            hook(module)


class _ImportHookFinder(abc.MetaPathFinder, set[str]):
    def find_spec(
        self,
        fullname: str,
        path: Sequence[str] | None,
        target: ModuleType | None = None,
        /,
    ) -> ModuleSpec | None:
        with _LOCK:
            if fullname not in _HOOKS or fullname in self:
                return None

        self.add(fullname)
        try:
            if (
                (spec := util.find_spec(fullname))
                and (loader := spec.loader)
                and not isinstance(loader, _ImportHookChainedLoader)
            ):
                spec.loader = _ImportHookChainedLoader(loader)
                return spec
        finally:
            self.remove(fullname)
        return None


def register_post_import_hook(hook: Callback[ModuleType], name: str) -> None:
    """Register a new post import hook for the target module name.

    This will result in a proxy callback being registered which will defer
    loading of the specified module containing the callback function until
    required.

    Simplified version of wrapt.register_post_import_hook.
    """
    with _LOCK:
        global _INITIALIZED  # noqa: PLW0603
        if not _INITIALIZED:
            _INITIALIZED = True
            sys.meta_path.insert(0, _ImportHookFinder())

        if (module := sys.modules.get(name)) is None:
            _HOOKS.setdefault(name, []).append(hook)
        else:
            hook(module)


def when_imported[H: Callback[ModuleType]](name: str) -> Callable[[H], H]:
    """Create decorator making a function a post import hook for a module.

    Simplified version of wrapt.when_imported.
    """

    def wrapper(hook: H) -> H:
        register_post_import_hook(hook, name)
        return hook

    return wrapper
